#include <fstream>
#include <iomanip>
#include <string.h> // for strdupa
//#include <stddef.h> // for strtok#include

#include "channelflow/flowfield.h"
#include "channelflow/symmetry.h"

using namespace std;

namespace channelflow {

FieldSymmetry::FieldSymmetry()
  :
  su_(1),
  sx_(1),
  sy_(1),
  sz_(1),
  ax_(0),
  az_(0)
{}


FieldSymmetry::FieldSymmetry(int su)
  :
  su_(su),
  sx_(1),
  sy_(1),
  sz_(1),
  ax_(0.0),
  az_(0.0)
{}

FieldSymmetry::FieldSymmetry(Real ax, Real az)
  :
  su_(1),
  sx_(1),
  sy_(1),
  sz_(1),
  ax_(ax),
  az_(az)
{}

FieldSymmetry::FieldSymmetry(int sx, int sy, int sz, Real ax, Real az, int su)
  :
  su_(su),
  sx_(sx),
  sy_(sy),
  sz_(sz),
  ax_(ax),
  az_(az)
{
  if (abs(su) != 1 || abs(sx) != 1 || abs(sy) != 1 || abs(sz) != 1 ) {
    cerr << "error in FieldSymmetry::FieldSymmetry(su, sx, sy, sz, ax, az) :" << endl;
    cerr << "su,sx,sy,sz should be +/-1, but they are "
	 << su << ',' << sx << ',' << sx << ',' << sz << endl;
    exit(1);
  }
}

FieldSymmetry::FieldSymmetry(bool sx, bool sy, bool sz, Real ax, Real az, bool su)
  :
  su_(su ? -1:1),
  sx_(sx ? -1:1),
  sy_(sy ? -1:1),
  sz_(sz ? -1:1),
  ax_(ax),
  az_(az)
{}


FieldSymmetry::FieldSymmetry(const string& filebase)
  :
  su_(1),
  sx_(1),
  sy_(1),
  sz_(1),
  ax_(0.0),
  az_(0.0)
{


  ifstream is;
  string filename = ifstreamOpen(is, filebase, ".asc");
  if (!is) {
    cerr << "FieldSymmetry::FieldSymmetry(filebase) : can't open file "
	 << filebase << " or " << (filebase+".asc") << endl;
    exit(1);
  }

  // Read in header. Form is "%N a b s"
  string comment;
  while (is.peek() == '%')
    getline(is, comment);
  is >> su_ >> sx_ >> sy_ >> sz_ >> ax_ >> az_;
  if (!is.good())
    cerr << "warning: bad istream in reading FieldSymmetry from file " << filename << endl;
  is.close();
}

int FieldSymmetry::s(int i) const {
  assert(i>=0 && i<3);
  int rtn = 0;
  switch (i) {
  case 0: rtn = sx_; break;
  case 1: rtn = sy_; break;
  case 2: rtn = sz_; break;
  default:
    cerr << "error in FieldSymmetry::s(int i) : i == " << i << " != 0,1, or 2" << endl;
    exit(1);
  }
  return rtn;
}

int FieldSymmetry::sign(int i) const {
  return s(i);
}

void FieldSymmetry::save(const std::string& filebase) const {
  string filename = appendSuffix(filebase, ".asc");
  ofstream os(filename.c_str());
  char s = ' ';
  os << setprecision(17);
  os << "% (s sx sy sz ax az) [u,v,w](x,y,z) = s [sx u, sy v, sz w](sx x + ax Lx, sy y, sz z + az Lz)\n";
  os << "% s sx sy sz ax az\n";

  os << su_ << s << sx_ << s << sy_ << s << sz_ << s << ax_ << s << az_ << endl;
}


FlowField FieldSymmetry::operator()(const FlowField& u) const {
  FlowField v(u);
  v *= (*this);
  return v;
}


RealProfile FieldSymmetry::operator()(const RealProfile &u) const {
  RealProfile v(u);
  v *= (*this);
  return v;
}

FieldSymmetry& FieldSymmetry::operator *= (Real c) {
  if (su_ != 1 || sx_ != 1 || sy_ != 1 || sz_ != 1)
    cerr << "warning: FieldSymmetry *= Real should only be applied to pure translations" << endl;
  ax_ = fmod(c*ax_+0.5, 1.0) - 0.5;
  az_ = fmod(c*az_+0.5, 1.0) - 0.5;
  return *this;
}

FieldSymmetry& FieldSymmetry::operator *= (const FieldSymmetry& p) {
  su_ *= p.su();
  sx_ *= p.sx();
  sy_ *= p.sy();
  sz_ *= p.sz();
  ax_ = fmod(p.ax() + p.sx()*ax_ + 0.5, 1.0) - 0.5;
  az_ = fmod(p.az() + p.sz()*az_ + 0.5, 1.0) - 0.5;
  return *this;
}

FieldSymmetry operator* (const FieldSymmetry& p, const FieldSymmetry& q) {
  int su = p.su() * q.su();
  int sx = p.sx() * q.sx();
  int sy = p.sy() * q.sy();
  int sz = p.sz() * q.sz();
  Real ax = fmod(p.ax() + p.sx()*q.ax()+0.5, 1.0) - 0.5;
  Real az = fmod(p.az() + p.sz()*q.az()+0.5, 1.0) - 0.5;
  return FieldSymmetry(sx,sy,sz,ax,az,su);
}

FieldSymmetry inverse(const FieldSymmetry& s) {
  return FieldSymmetry(s.sx(), s.sy(), s.sz(), -s.ax(), -s.az(), s.su());
}

FlowField operator*(const FieldSymmetry& s, const FlowField& u) {
  return s(u);
}

bool operator == (const FieldSymmetry& p, const FieldSymmetry& q) {
  return (p.su() == q.su() &&
	  p.sx() == q.sx() &&
	  p.sy() == q.sy() &&
	  p.sz() == q.sz() &&
	  p.ax() == q.ax() &&
	  p.az() == q.az());
}

bool operator != (const FieldSymmetry& p, const FieldSymmetry& q) {
  return !(p == q);
}

ostream& operator<<(ostream& os, const FieldSymmetry& s) {
  streamsize p = os.precision();
  os << '('
     << setw(3) << setiosflags(ios::right) << s.su()
     << setw(3) << setiosflags(ios::right) << s.sx()
     << setw(3) << setiosflags(ios::right) << s.sy()
     << setw(3) << setiosflags(ios::right) << s.sz() << "   "
     << resetiosflags(ios::adjustfield)
     << setw(p+8) << setiosflags(ios::left) << s.ax() << ' '
     << setw(p+8) << setiosflags(ios::left) << s.az()
     << ')' << resetiosflags(ios::adjustfield);
  return os;
}

istream& operator>>(istream& is, FieldSymmetry& s) {
  char c;
  is >> c;
  if (c !='(' ) {
    cerr << "error in istream >> FieldSymmetry : expected '(' as opening character" << endl;
    cerr << "got '" << c << "' instead" << endl;
  }
  int su,sx,sy,sz;
  Real ax,az;
  is >> sx >> sy >> sz >> ax >> az >> su >> c;
  if (c !=')' )
    cferror("error in istream >> FieldSymmetry : expected ')' as closing character");

  s = FieldSymmetry(sx,sy,sz,ax,az,su);
  return is;
}

SymmetryList::SymmetryList()
{}

SymmetryList::SymmetryList(int N)
  :
  array<FieldSymmetry>(N)
{}

SymmetryList::SymmetryList(const string& filebase) {

  ifstream is;
  ifstreamOpen(is, filebase, ".asc");
  if (!is) {
    cerr << "SymetryList::SymmetryList(filebase) : can't open file "
	 << filebase << " or " << (filebase+".asc") << endl;
    exit(1);
  }
  int N;
  is >> N;
  resize(N);
  for (int n=0; n<N; ++n)
    is >> (*this)[n];
}

void SymmetryList::save(const string& filebase) const {
  string filename = appendSuffix(filebase, ".symms");
  ofstream os(filename.c_str());

  int N = this->length();
  os << N << '\n';
  os << setprecision(17);
  for (int n=0; n<N; ++n)
    os << (*this)[n] << '\n';
}


} // namespace channelflow
