// License declaration at end of file

// Program for writing HDF5 files from flowfield data, essentially a
// modification of field2ascii. Some of the code and the accopmanying
// comments are taken from the HDF group website (www.hdfgroup.org).

#include <iostream>
#include <fstream>
#include <iomanip>
#include <config.h>
#include "channelflow/flowfield.h"
#include "channelflow/dns.h"
#include "channelflow/utilfuncs.h"


#ifndef HAVE_LIBHDF5_CPP
int main() {
  std::cout << "field2hdf5 : Sorry, this functionality is only supported if you have the\n";
  std::cout << "installed. Please install the HDF5 libraries, recompile channelflow, and\n";
  std::cout << "try again.\n";
}
#else

#include "H5Cpp.h"  // header file for HDF5
//#include "time.h"

using namespace std;
using namespace channelflow;
using namespace H5;

void write_attr(H5File& file, const string& name,  double value);
void write_attr(H5File& file, const string& name,  int value);
void write_1d_vec(const Vector& vec_in, H5File& file, const string& name);
void write_4d_var(const FlowField& u, H5File& file, const string& name);

int main(int argc, char* argv[]) {

  string purpose("save FlowField as HDF5 file, containing:\n"
		 "\t  - the geometry arrays x,y,z in the group /geom \n"
		 "\t  - the components of the field in the group /data \n"
		 "\t  - the values of Nx, Ny, Nz, Lx, a, and b as attributes of the root group   \n"
		 "\tFields with 2/3-style padding for dealiasing are saved on reduced grids (Nx,Nz -> 2Nx/3, 2Nz/3).");

  ArgList args(argc, argv, purpose);

  const bool padding = args.getflag("-p", "--padding", "save full Nx x Ny x Nz grid rather eliminating padded modes");
  const string uname = args.getstr(2, "<fieldname>",   "flowfield input file");
  const string afile = args.getstr(1, "<hdf5file>",    "hdf5 output file");

  args.check();

  FlowField u(uname);

  if (!padding && u.dealiased()) {
    int Nx = (2*u.Nx())/3;
    int Nz = (2*u.Nz())/3;
    FlowField v(Nx,u.Ny(),Nz,u.Nd(),u.Lx(),u.Lz(),u.a(), u.b());
    v.interpolate(u);
    u = v;
  }

  u.makeState(Physical, Physical);

  H5std_string hdf5fname = afile+".h5";

  // create HDF5 file
  //clock_t start;
  //clock_t stop;
  //int N=100;
  //start = clock();

  H5File file(hdf5fname,H5F_ACC_TRUNC);

  // create the groups we will need
  file.createGroup( "/geom" );
  file.createGroup( "/data" );

  // write the field
  write_4d_var(u,file,"/data/u");

  // write the 1-D grid vectors
  write_1d_vec(u.xgridpts(), file, "/geom/x");
  write_1d_vec(u.ygridpts(), file, "/geom/y");
  write_1d_vec(u.zgridpts(), file, "/geom/z");

  // Write other information as attributes
  write_attr(file,"Nx",u.Nx());
  write_attr(file,"Ny",u.Ny());
  write_attr(file,"Nz",u.Nz());
  write_attr(file,"Lx",u.Lx());
  write_attr(file,"Lz",u.Lz());
  write_attr(file,"a", u.a()); // ymin
  write_attr(file,"b", u.b()); // ymax

  //stop = clock();
  //cout << "hdf5   write time == " << Real(stop-start)/CLOCKS_PER_SEC << endl;


  //start = clock();
  u.binarySave("foo");
  //stop = clock();
  //cout << "chflow write time == " << Real(stop-start)/CLOCKS_PER_SEC << endl;
}


// a function for writing a 4-D double precision array
void write_4d_var(const FlowField& u, H5File& file, const string& name) {

  const int Nx = u.Nx();
  const int Ny = u.Ny();
  const int Nz = u.Nz();
  const int Nd = u.Nd();
  //cout << Nx << "\n";
  //cout << Ny << "\n";
  //cout << Nz << "\n";
  //cout << Nd << "\n";

  double* data = new double[Nx*Ny*Nz*Nd]; // create array of data to be written

  for (int nx=0; nx<Nx; ++nx)
    for (int ny=0; ny<Ny; ++ny)
      for (int nz=0; nz<Nz; ++nz)
	for (int i=0; i<Nd; ++i)
	  data[i + Nd*(nz + Nz*(ny + Ny*nx))] = u(nx,ny,nz,i);

  //cout <<"data done \n";
  const hsize_t     dimsf[4] = {Nd,Nx,Ny,Nz};   // dataset dimensions
  const int dsize=Nx*Ny*Nz*Nd;

  herr_t status;
  hid_t natflt(H5T_NATIVE_DOUBLE); // datatypes
  hid_t dtypet(H5T_IEEE_F64BE);
  FloatType write_dtype( PredType:: IEEE_F64BE  );

  DataSpace dataspace4d( 4, dimsf );  // create dataspace
  DataSet dset = file.createDataSet(name.c_str(), write_dtype, dataspace4d ); // create dataset

  status = H5Tconvert(natflt,dtypet,dsize,data,NULL,H5P_DEFAULT); // convert to 64bit big-endian
  if (status <0) {
    cout << "HDF5 datatype conversion failed! Status: " << status;
  }
  dset.write( data, write_dtype  );  // write data to dataset
  dset.close();
  delete data;
}

// a function for writing a 1-D double precision vector
void write_1d_vec(const Vector& vec_in, H5File& file, const string& name) {

  const int Nx=vec_in.length();
  double* vec1d = new double[Nx];
  for (int nx=0; nx<Nx; ++nx)  // create an array, can't write a vector directly
    vec1d[nx] = vec_in(nx);

  const hsize_t     hnx[1]={Nx};
  const int dsize=Nx;

  herr_t status;
  hid_t natflt(H5T_NATIVE_DOUBLE); // datatypes
  hid_t dtypet(H5T_IEEE_F64BE);
  FloatType write_dtype(PredType:: IEEE_F64BE);

  DataSpace dataspace1d( 1, hnx ); // create dataspace
  DataSet dset = file.createDataSet(name.c_str(), write_dtype, dataspace1d ); // create dataset
  status = H5Tconvert(natflt,dtypet,dsize,vec1d,NULL,H5P_DEFAULT); // convert to 64bit big-endian
  if (status <0) {
    cout << "HDF5 datatype conversion failed! Status: " << status;
  }
  dset.write( vec1d, write_dtype );  // write data to dataset
  dset.close();
  delete vec1d;
}

// a function for writing integer attributes
void write_attr(H5File& file, const string& name,  int value) {

  herr_t status;
  hid_t natint(H5T_NATIVE_INT);
  hid_t inttypet(H5T_STD_I32BE);
  IntType write_dtype(PredType:: STD_I32BE);

  status = H5Tconvert(natint,inttypet,1,&value,NULL,H5P_DEFAULT);
  if (status <0) {
    cout << "HDF5 datatype conversion failed! Status: " << status;
  }

  Group rootgr = file.openGroup("/");
  DataSpace attr_dataspace = DataSpace( H5S_SCALAR );
  Attribute attr = rootgr.createAttribute(name.c_str(), write_dtype, attr_dataspace);
  attr.write(write_dtype, &value);
}

// overloaded function for writing double attributes
void write_attr(H5File& file, const string& name,  double value) {

  herr_t status;
  hid_t natflt(H5T_NATIVE_DOUBLE); //datatypes
  hid_t dtypet(H5T_IEEE_F64BE);
  FloatType write_dtype(PredType:: IEEE_F64BE);

  status = H5Tconvert(natflt,dtypet,1,&value,NULL,H5P_DEFAULT);
  if (status <0) {
    cout << "HDF5 datatype conversion failed! Status: " << status;
  }

  Group rootgr = file.openGroup("/");
  DataSpace attr_dataspace = DataSpace( H5S_SCALAR );
  Attribute attr = rootgr.createAttribute(name.c_str(),write_dtype , attr_dataspace);
  attr.write(write_dtype, &value);
}
#endif

/* couette.cpp: time-integration class for spectral Navier-Stokes simulator
 * channelflow-1.3
 *
 * Copyright (C) 2001-2009  Milos Ilak and John F. Gibson
 *
 * gibson@cns.physics.gatech.edu  johnfgibson@gmail.com
 *
 * Center for Nonlinear Science
 * School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 * 404 385 2509
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation version 2
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */
